# encoding: utf-8

# -------------------------------------------------------------------------------
# (C) Copyright 2019 - 2022 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates
# and licensors (“Micro Focus”) are as may be set forth in the express warranty
# statements accompanying such products and services. Nothing herein should be
# construed as constituting an additional warranty. Micro Focus shall not be
# liable for technical or editorial errors or omissions contained herein. The
# information contained herein is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains confidential
# information and a valid license is required for possession, use or copying. If
# this work is provided to the U.S. Government, consistent with FAR 12.211 and
# 12.212, Commercial Computer Software, Computer Software Documentation, and
# Technical Data for Commercial Items  are licensed to the U.S. Government under
# vendor's standard commercial license.
# -------------------------------------------------------------------------------

# File:	modules/Nss.ycp
# Package:	Configuration of nss
# Summary:	Nss settings, input and output functions
# Authors:	Mark Clark <mclark@novell.com>
#
# $Id: Nss.ycp,v 1.1.2.73 2005/08/03 04:15:25 mclark Exp $
#
# Representation of the configuration of nss.
# Input and output routines.
require "yast"

module Yast
  class NssClass < Module
    def main
      Yast.import "UI"
      textdomain "novell-nss"

      Yast.import "Popup"
      Yast.import "Mode"
      Yast.import "Stage"
      Yast.import "Progress"
      Yast.import "Report"
      Yast.import "Summary"
      Yast.import "NovellUtils"
      Yast.import "Edirectory"
      Yast.import "EdirUtils"
      Yast.import "NovellLum"
      Yast.import "Misc"
      Yast.import "String"
      Yast.import "Service"
      Yast.import "NovellLDAP"
      Yast.import "NovellPatterns"
      Yast.import "NovellAddonInstall"
      Yast.import "Wizard"

      # Settings Map
      @auto_clone = false
      @Settings = {}
      @Settings_nit = {}
      @oldSettings = {} # OES 1.0 settings for upgrade
      @edir_service_configured = false
      @schemaLogPath = "/var/opt/novell/eDirectory/log/oes_schema.log"
      @oesContext = "ou=OESSystemObjects"
      @userChangedData = false
      @doRecalculate = true

      @xad_packages_installed = false
      @xadMode = false

      @xadTestMode = false

      # Data was modified?
      @modified = false

      @upgrade_from = ""
      @proposal_valid = false

      # Are we in reinstall mode?
      @reinstallMode = false

      # Write only, used during autoinstallation.
      # Don't run services and SuSEconfig, it's all done at one place.
      @write_only = false
      @globals_configured = false

      # Abort function
      # return boolean return true if abort
      @AbortFunction = fun_ref(method(:Modified), "boolean ()")

      @scr_conf_path = path(".sysconfig.nss")
    end

    # Abort function
    # @return [Boolean] return true if abort
    def Abort
      return @AbortFunction.call == true if @AbortFunction != nil
      false
    end

    # Data was modified?
    # @return true if modified
    def Modified
      Builtins.y2debug("modified=%1", @modified)
      @modified
    end
    def ReadOption(optname, replace)
      optvalue = Convert.to_string(
        SCR.Read(Ops.add(@scr_conf_path, Builtins.topath(optname)))
      )

      optvalue = "" if optvalue == nil
      optvalue = replace if optvalue == "" && replace != nil

      optvalue
    end

    def GetNameMap
      nameMap = Ops.get_string(
        Edirectory.EdirSettings,
        "XAD_CONVERT_EXISTING_CONTAINER",
        ""
      )
      Builtins.y2milestone(
        "nss.GetXadObjects:are we name mapped from eDir map:%1",
        nameMap
      )
      if nameMap == nil || nameMap == ""
        nameMap = Convert.to_string(
          SCR.Read(path(".sysconfig.novell.xad.XAD_CONVERT_EXISTING_CONTAINER"))
        ) # next try reading the xad conf file
        Builtins.y2milestone(
          "nss.GetXadUnixObjects:are we name mapped from xad conf file:%1",
          nameMap
        )
      end
      nameMap = "no" if nameMap == nil || nameMap == ""

      nameMap
    end

    def GetNameMapContainer
      nameMapContainer = Ops.get_string(
        Edirectory.EdirSettings,
        "XAD_EXISTING_CONTAINER",
        ""
      )
      Builtins.y2milestone(
        "nss.GetNameMapContainer:name mapped container from eDir map:%1",
        nameMapContainer
      )
      if nameMapContainer == nil || nameMapContainer == ""
        nameMapContainer = Convert.to_string(
          SCR.Read(path(".sysconfig.novell.xad.XAD_EXISTING_CONTAINER"))
        ) # next try reading the xad conf file
        Builtins.y2milestone(
          "nss.GetXadUnixObjects:name mapped container from xad conf file:%1",
          nameMapContainer
        )
      end
      if nameMapContainer == nil || nameMapContainer == ""
        nameMapContainer = ""
      else
        nameMapContainer = NovellUtils.Nds2Ldap(nameMapContainer)
      end

      nameMapContainer
    end

    def GetOrCreateGID(group)
      execute_cmd = ""
      ret = -1
      gid = ""

      local_gid = NovellUtils.Get_GID(group)
      if local_gid == nil || local_gid == ""
        gid = " -g 81" if group == "novlxtier"

        execute_cmd = Ops.add(Ops.add(Ops.add("groupadd", gid), " -r "), group)
        ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
        local_gid = NovellUtils.Get_GID(group)
        Builtins.y2milestone(
          "Nss:Executed: %1, ret is %2, GID is %3",
          execute_cmd,
          ret,
          local_gid
        )
      end
      local_gid
    end

    def GetOrCreateUID(user, group)
      execute_cmd = ""
      ret = -1
      uid = ""

      local_uid = NovellUtils.Get_UID(user)
      if local_uid == nil || local_uid == ""
        if user == "novlxregd"
          uid = " -u 81"
        elsif user == "novlxsrvd"
          uid = " -u 82"
        end
        execute_cmd = Ops.add(
          Ops.add(
            Ops.add(Ops.add("useradd -s /bin/false -r -g ", group), uid),
            " "
          ),
          user
        )
        ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
        local_uid = NovellUtils.Get_UID(user)
        Builtins.y2milestone(
          "Nss:Executed: %1, ret is %2, UID is %3",
          execute_cmd,
          ret,
          local_uid
        )
      end
      local_uid
    end

    def ShutDownNamcd
      # Because groupdel see the LUM users, we need to shut down namcd while we delete the local group.
#      execute_cmd = "rcnamcd stop"
#      ret = 0
      if @xadTestMode == false
 #       ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
	Service.Stop("namcd")
      end
 #     Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)

      true # fix this.
    end

    def DeleteXtierLocals
      execute_cmd = "userdel novlxregd"
      ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
      Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)

      execute_cmd = "userdel novlxsrvd"
      ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
      Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)

      execute_cmd = "groupdel novlxtier"
      ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
      Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)
      if ret != 0
        Builtins.y2milestone(
          "Nss:Executed: %1, ret is %2. Delaying and trying again.",
          execute_cmd,
          ret
        )
        Builtins.sleep(6000)
        ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
        Builtins.y2milestone("Nss:Executed: %1, ret is %2.", execute_cmd, ret)
      end

      true
    end

    def DeleteLocals
      # In case of upgrade from OES 1.0
      execute_cmd = "userdel novlwww"
      ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
      Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)

      execute_cmd = "userdel wwwrun"
      ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
      Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)

      execute_cmd = "groupdel www"
      ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
      Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)
      if ret != 0
        Builtins.y2milestone(
          "Nss:Executed: %1, ret is %2. Delaying and trying again.",
          execute_cmd,
          ret
        )
        Builtins.sleep(6000)
        ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
        Builtins.y2milestone("Nss:Executed: %1, ret is %2.", execute_cmd, ret)
      end

      true
    end

    def GetGIDfromLUM(group)
      rmap = {}
      retVal = ""
      idx = nil

      execute_cmd = Ops.add(Ops.add("namgrouplist ", group), " | cut -d':' -f3")
      Builtins.y2milestone("Nss:GetGIDfromLUM:execute_cmd: %1", execute_cmd)

      # retry loop
      Builtins.foreach([1, 2, 3]) do |i|
        rmap = Convert.to_map(
          SCR.Execute(path(".target.bash_output"), execute_cmd)
        )
        if Ops.get_integer(rmap, "exit", -1) != 0
          Builtins.y2milestone(
            "Nss:GetGIDfromLUM:Error getting group ID for group %1 from LUM",
            group
          )
          next ""
        end
        retVal = NovellUtils.CleanUpString(Ops.get_string(rmap, "stdout", ""))
        idx = Builtins.search(retVal, "No such object")
        if idx == nil
          raise Break
        else
          retVal = Builtins.sformat(
            "Replace this string with GID returned from command: namgrouplist %1",
            group
          )
        end
        Builtins.sleep(15000) if Ops.less_than(i, 3) #sleep 15 seconds
        Builtins.y2milestone(
          "Nss:GetGIDfromLUM:Error getting group ID for group %1 from LUM (No such object exists) Retry: %2",
          group,
          i
        )
      end

      retVal
    end

    def GetUIDfromLUM(user)
      rmap = {}
      retVal = ""
      idx = nil

      execute_cmd = Ops.add(Ops.add("namuserlist ", user), " | cut -d':' -f3")
      Builtins.y2milestone("Nss:GetUIDfromLUM:execute_cmd: %1", execute_cmd)

      # retry loop
      Builtins.foreach([1, 2, 3]) do |i|
        rmap = Convert.to_map(
          SCR.Execute(path(".target.bash_output"), execute_cmd)
        )
        if Ops.get_integer(rmap, "exit", -1) != 0
          Builtins.y2milestone(
            "Nss:GetUIDfromLUM:Error getting user ID for user %1 from LUM",
            user
          )
          next ""
        end
        retVal = NovellUtils.CleanUpString(Ops.get_string(rmap, "stdout", ""))
        idx = Builtins.search(retVal, "No such object")
        if idx == nil
          raise Break
        else
          retVal = Builtins.sformat(
            "Replace this string with UID returned from command: namuserlist %1",
            user
          )
        end
        Builtins.sleep(15000) if Ops.less_than(i, 3) #sleep 15 seconds
        Builtins.y2milestone(
          "Nss:GetUIDfromLUM:Error getting user ID for user %1 from LUM (No such object exists) Retry: %2",
          user,
          i
        )
      end

      retVal
    end

    # Generate a script line to Find all files with a given group and user ID and change to a new group and user ID.
    def GenerateScriptForFindFilesAndChangeGIDandUID(old_uid, old_gid, new_uid, new_gid)
      if old_uid == nil || old_uid == "" || old_gid == nil || old_gid == "" ||
          new_uid == nil ||
          new_uid == "" ||
          new_gid == nil ||
          new_gid == ""
        Builtins.y2milestone(
          "Nss:GenerateScriptForFindFilesAndChangeGIDandUID:Empty parameter."
        )
      end

      write_cmd = ""
      if old_uid != new_uid || old_gid != new_gid
        write_cmd = Ops.add(
          Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      "find / -follow -printf " + "\"%p\\n\"" +
                        " 2>/dev/null | xargs chown --from ",
                      old_uid
                    ),
                    ":"
                  ),
                  old_gid
                ),
                " "
              ),
              new_uid
            ),
            ":"
          ),
          new_gid
        )
      end

      Builtins.y2milestone(
        "Nss:GenerateScriptForFindFilesAndChangeGIDandUID:Generating: %1",
        write_cmd
      )

      write_cmd
    end

    # Write out a script file to Find all files with a given group and user ID and change to a new group and user ID.
    def WriteScriptFileForChown(old_www_gid, new_www_gid, old_wwwrun_uid, new_wwwrun_uid)
      old_www_gid = "" if old_www_gid == nil
      new_www_gid = "" if new_www_gid == nil
      old_wwwrun_uid = "" if old_wwwrun_uid == nil
      new_wwwrun_uid = "" if new_wwwrun_uid == nil

      line1 = GenerateScriptForFindFilesAndChangeGIDandUID(
        old_wwwrun_uid,
        old_www_gid,
        new_wwwrun_uid,
        new_www_gid
      )

      if line1 == ""
        Builtins.y2milestone("Nss:no lines to write to nssid.sh")
        return true
      end

      # Start with an empty file.
      ret = SCR.Write(
        path(".target.string"),
        "/opt/novell/oes-install/nssid.sh",
        ""
      )

      # Write the two lines out.
      ret = SCR.Write(
        path(".target.string"),
        "/opt/novell/oes-install/nssid.sh",
        Ops.add(line1, "\n")
      )
      Builtins.y2milestone(
        "Nss:WriteScriptFileForChown:wrote script to the file /opt/novell/oes-install/nssid.sh"
      )

      # Flush the buffer.
      ret = SCR.Write(
        path(".target.string"),
        "/opt/novell/oes-install/nssid.sh",
        nil
      )

      true
    end


    # Append xtier chown commands to the nssid script file that must be manually run by the administrator after the install.
    def AppendScriptFileForXtierChown(local_novlxregd_uid, local_novlxtier_gid, lum_novlxregdUID, lum_novlxtierGID, local_novlxsrvd_uid, lum_novlxsrvdUID)
      line1 = GenerateScriptForFindFilesAndChangeGIDandUID(
        local_novlxregd_uid,
        local_novlxtier_gid,
        lum_novlxregdUID,
        lum_novlxtierGID
      )
      line2 = GenerateScriptForFindFilesAndChangeGIDandUID(
        local_novlxsrvd_uid,
        local_novlxtier_gid,
        lum_novlxsrvdUID,
        lum_novlxtierGID
      )

      if line1 == "" && line2 == ""
        Builtins.y2milestone("Nss:xtier:no lines to write to nssid.sh")
        return true
      end

      fileData = ""

      # First, see if the file exists..
      if Ops.greater_than(
          SCR.Read(path(".target.size"), "/opt/novell/oes-install/nssid.sh"),
          0
        )
        fileData = Convert.to_string(
          SCR.Read(path(".target.string"), "/opt/novell/oes-install/nssid.sh")
        )
      else
        SCR.Write(
          path(".target.string"),
          "/opt/novell/oes-install/nssid.sh",
          ""
        )
      end

      # Append the two lines to the end of the file.
      ret = SCR.Write(
        path(".target.string"),
        "/opt/novell/oes-install/nssid.sh",
        Ops.add(
          Ops.add(Ops.add(Ops.add(Ops.add(fileData, "\n"), line1), "\n"), line2),
          "\n"
        )
      )
      Builtins.y2milestone(
        "nss:AppendScriptFileForXtierChown:wrote script to the file /opt/novell/oes-install/nssid.sh"
      )

      # Flush the buffer.
      ret = SCR.Write(
        path(".target.string"),
        "/opt/novell/oes-install/nssid.sh",
        nil
      )

      true
    end



    # Get the work context value for the nssAdmin
    # @return work context
    def CalculateAndSetServerContext
      wrkContext = ""
      if Mode.installation && Stage.normal && @edir_service_configured
        wrkContext = Misc.SysconfigRead(
          path(".sysconfig.edirectory.CONFIG_EDIR_SERVER_CONTEXT"),
          ""
        )
        Builtins.y2milestone(
          "Nss:Read (post install mode).eDir server context from conf file: %1",
          wrkContext
        )
      else
        wrkContext = Ops.get_string(
          Edirectory.EdirSettings,
          "CONFIG_EDIR_SERVER_CONTEXT",
          ""
        )
        Builtins.y2milestone(
          "Nss:Read.eDir server context from map: %1",
          wrkContext
        )
        if wrkContext == nil || wrkContext == ""
          wrkContext = Misc.SysconfigRead(
            path(".sysconfig.edirectory.CONFIG_EDIR_SERVER_CONTEXT"),
            ""
          )
          Builtins.y2milestone(
            "Nss:Read.eDir server context from conf file: %1",
            wrkContext
          )
        end
      end
      Ops.set(@Settings, "CONFIG_NSS_EDIR_SERVER_CONTEXT", wrkContext)

      wrkContext
    end

    # Get and Set NSS context values
    # @return true on success
    def CalculateAndSetNssAdminAndServerContext
      wrkContext = CalculateAndSetServerContext()
      if wrkContext == nil || wrkContext == ""
        wrkContext = NovellUtils.GetContextNDS(
          Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_ADMIN_CONTEXT", "")
        )
        Builtins.y2milestone(
          "Nss:Read.admin user context (could not get eDir server context): %1",
          wrkContext
        )
        wrkContext = "" if wrkContext == nil || wrkContext == ""
      end

      if @xadMode
        nameMap = GetNameMap()
        nameMapContext = GetNameMapContainer()

        if nameMap == "yes"
          wrkContext = Ops.add(
            Ops.add(@oesContext, "."),
            NovellUtils.Ldap2Nds(nameMapContext)
          )
        else
          wrkContext = NovellUtils.Ldap2Nds(wrkContext)
        end
      end
      Ops.set(
        @Settings,
        "CONFIG_NSS_NSSADMIN_DN",
        Ops.add(
          Ops.add(Ops.add("cn=", NovellUtils.GetHostname), "admin."),
          NovellUtils.Ldap2Nds(wrkContext)
        )
      )

      true
    end

    # NitRangeDialog function
    # @return widget
    def NitRangeDialog
      start_range = Ops.get_integer(
        @Settings_nit,
        "CONFIG_NIT_START_RANGE",
        100000
      )
      end_range = Ops.get_integer(@Settings_nit, "CONFIG_NIT_END_RANGE", 200000)
      uid_start_range = 100000
      uid_end_range = 2147483647

      contents = VBox(
        #           `Left(`Label( _("Novell Identity Translator Configuration"))),
        # frame label
        Frame(
          _("&UID Range"),
          HBox(
            # int field label
            IntField(
              Id(:start_range),
              _("&Start"),
              uid_start_range,
              Ops.subtract(uid_end_range, 1),
              start_range
            ),
            # int field label
            IntField(
              Id(:end_range),
              _("&End"),
              uid_start_range,
              uid_end_range,
              end_range
            )
          )
        )
      )

      deep_copy(contents)
    end

    # NitRangeNextButton function
    # @return start_range and end_range if true, else `continue

    def NitRangeNextButton
      start_range = Convert.to_integer(UI.QueryWidget(Id(:start_range), :Value))
      end_range = Convert.to_integer(UI.QueryWidget(Id(:end_range), :Value))
      if Ops.greater_or_equal(start_range, end_range)
        Report.Error(
          Builtins.sformat(
            _("UID start range should be lesser than the UID end range.")
          )
        )
        return :continue
      else
        Ops.set(@Settings_nit, "CONFIG_NIT_START_RANGE", start_range)
        Ops.set(@Settings_nit, "CONFIG_NIT_END_RANGE", end_range)
      end

      nil
    end

    # NitReadSettings function
    # reads uid_start_range and uid_end_range
    def NitReadSettings
      cmd = Builtins.sformat(
        "/usr/bin/nitconfig get uid-start-range |cut -d\"=\" -f2 |awk '{print $1}'"
      )
      val = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))

      start_range = Builtins.tointeger(
        NovellUtils.CleanUpString(Ops.get_string(val, "stdout", ""))
      )

      cmd = Builtins.sformat(
        "/usr/bin/nitconfig get uid-end-range |cut -d\"=\" -f2 |awk '{print $1}'"
      )
      val = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
      end_range = Builtins.tointeger(
        NovellUtils.CleanUpString(Ops.get_string(val, "stdout", ""))
      )
      Ops.set(@Settings_nit, "CONFIG_NIT_START_RANGE", start_range)
      Ops.set(@Settings_nit, "CONFIG_NIT_END_RANGE", end_range)
      Builtins.y2milestone(
        "uid_start_range %1, uid_end_range %2",
        start_range,
        end_range
      )

      nil
    end

    # NitRangeSettings function
    # sets uid_start_range and uid_end_range
    def NitRangeSettings
      Builtins.y2milestone("Seeting Nit Range")
      cmd = Builtins.sformat(
        "/usr/bin/nitconfig set uid-start-range=%1",
        Ops.get_integer(@Settings_nit, "CONFIG_NIT_START_RANGE", 100000)
      )
      val = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
      Builtins.y2milestone("cmd executed is %1", cmd)
      Builtins.y2milestone("val is %1", val)

      cmd = Builtins.sformat(
        "/usr/bin/nitconfig set uid-end-range=%1",
        Ops.get_integer(@Settings_nit, "CONFIG_NIT_END_RANGE", 200000)
      )
      val = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))

      Builtins.y2milestone("cmd executed is %1", cmd)
      Builtins.y2milestone("val is %1", val)

      nil
    end

    # Read all nss settings
    # @return true on success
    def Read
      # Nss read dialog caption
      caption = _("Initializing OES Storage Services Configuration")

      # TODO FIXME Set the right number of stages
      steps = 2

      sl = 500
      Builtins.sleep(sl)

      # TODO FIXME Names of real stages
      # We do not set help text here, because it was set outside
      Progress.New(
        caption,
        " ",
        steps,
        [
          # Progress stage 2/3
          _("Read the prior OES Storage Services settings"),
          _("Checking dependencies")
        ],
        [
          # Progress step 2/3
          _("Reading the previous settings..."),
          _("Checking dependencies..."),
          # Progress finished
          _("Finished")
        ],
        ""
      )

      # read current settings
      return false if Abort()
      Progress.NextStage

      # Test for the upgrade scenario
      upgradeInfo = {}
      # See if we are doing an upgrade from OES 2 sp2

      Ops.set(@Settings, "CONFIG_NSS_EDIR_SERVER_CONTEXT", "")

      @upgrade_from = NovellUtils.UpgradeIsFrom
      if (Mode.update || Mode.autoupgrade) &&  @upgrade_from != ""
        @oldSettings = NovellUtils.GetOldSysconfigMap("nss")
      end

      Builtins.y2milestone(
        "nss:
        upgradeFrom = #{@upgrade_from}"
      )
      Builtins.y2milestone("nss:oldSettings: %1", @oldSettings)

      info = NovellAddonInstall.GetProductInfo("edirectory")
      if Ops.get_boolean(info, "SERVICE_CONFIGURED", false) == true
        @edir_service_configured = true
      else
        @edir_service_configured = false
      end

      # First, get the default LDAP/eDir data
      progress_orig = Progress.set(false)
      NovellLDAP.Read
      Progress.set(progress_orig)

      setDefaultIP = false

      # See if XAD is installed. If so, default the data.
      @xad_packages_installed = NovellPatterns.IsPatternInstalled(
        "novell-domain-services"
      )
      Builtins.y2milestone(
        "Nss:xad_packages_installed:%1",
        @xad_packages_installed
      )
      if @xad_packages_installed
        if Ops.get_string(Edirectory.EdirSettings, "XAD_SERVICE_CONFIGURED", "") == "yes" ||
            Convert.to_string(
              SCR.Read(path(".sysconfig.novell.xad.SERVICE_CONFIGURED"))
            ) == "yes"
          @xadMode = true
        end
      end

      @xadMode = true if @xadTestMode

      Builtins.y2milestone("Nss:xadMode:%1", @xadMode)

      if @xadMode
        change_oes_context = Builtins.tolower(
          NovellUtils.ReadOption("CHANGED_OES_CONTEXT")
        )
        Builtins.y2milestone(
          "Nss:dsfw context changed is %1",
          change_oes_context
        )
        @oesContext = "ou=novell" if change_oes_context == "yes"
      end

      # determine if NSS configuration already exists
      service_configured = Builtins.tolower(
        ReadOption("SERVICE_CONFIGURED", "no")
      )
      service_previously_configured = Builtins.tolower(
        ReadOption("SERVICE_PREVIOUSLY_CONFIGURED", "no")
      )

      if Mode.update || Mode.autoupgrade
        # Fix for bug #680935
        Ops.set(
          @Settings,
          "CONFIG_NSS_LDAP_SERVER",
          Ops.get_string(@oldSettings, "CONFIG_NSS_LDAP_SERVER", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_NSS_NSSADMIN_DN",
          Ops.get_string(@oldSettings, "CONFIG_NSS_NSSADMIN_DN", "")
        )
         Ops.set(
          @Settings,
          "CONFIG_NSS_EDIR_SERVER_CONTEXT",
          CalculateAndSetServerContext()
        )

        @doRecalculate = false # don't recalculate these values in Write() since they are read from old sysconfig file.
      else
        if service_configured == "yes"
          @reinstallMode = true
          # SPECIFIC SETTINGS EXIST
          Ops.set(
            @Settings,
            "CONFIG_NSS_LDAP_SERVER",
            ReadOption("CONFIG_NSS_LDAP_SERVER", "")
          )
          Ops.set(
            @Settings,
            "CONFIG_NSS_NSSADMIN_DN",
            ReadOption("CONFIG_NSS_NSSADMIN_DN", "")
          )

          CalculateAndSetServerContext()
          @doRecalculate = false # don't recalculate these values in Write() since they are read from the existing sysconfig file.
        else
          if service_previously_configured != "yes"
            @reinstallMode = false
          else
            @reinstallMode = true
          end

          setDefaultIP = true

          bVal = CalculateAndSetNssAdminAndServerContext()

          wrkContext = ""
          server_list = NovellLDAP.GetLocalServers(nil)
          if Ops.greater_than(Builtins.size(server_list), 0)
            Ops.set(
              @Settings,
              "CONFIG_NSS_LDAP_SERVER",
              Ops.get_string(
                Ops.get(server_list, 0, {}),
                "CONFIG_LDAP_SERVER_IP",
                ""
              )
            )
          else
            server_list = NovellLDAP.GetRemoteServers(nil)
            Ops.set(
              @Settings,
              "CONFIG_NSS_LDAP_SERVER",
              Ops.get_string(
                Ops.get(server_list, 0, {}),
                "CONFIG_LDAP_SERVER_IP",
                ""
              )
            )
          end
        end
      end

      NitReadSettings()
      Builtins.y2debug("NSS READ Complete: %1", @Settings)
      @globals_configured = true

      # check dependencies
      return false if Abort()
      Progress.NextStage

      #    if (Mode::normal())
      #    {
      #        if (! NovellPatterns::EnsurePatternInstalled("novell-nss") )
      #        {
      #           return false;
      #        }
      #    }

      return false if Abort()
      Progress.NextStage
      Progress.Title(_("Finished"))
      Builtins.sleep(sl)

      return false if Abort()
      @modified = false
      true
    end

    def ChangeModForNssidFile
      execute_cmd = "chmod 755 /opt/novell/oes-install/nssid.sh"
      ret = Convert.to_integer(SCR.Execute(path(".target.bash"), execute_cmd))
      Builtins.y2milestone("Nss:Executed: %1, ret is %2", execute_cmd, ret)

      true
    end

    def DeleteLumUsers(lumUsers)
      lumUsers = deep_copy(lumUsers)
      bRet = false

      passwdFile = "/etc/passwd"

      # Backup the /etc/passwd file
      SCR.Execute(path(".target.bash"), "cp /etc/passwd /etc/passwd_OESsave")

      # Read the file into memory
      passwdData = Convert.to_string(
        SCR.Read(path(".target.string"), passwdFile)
      )
      buildStr = ""

      if passwdData != nil && passwdData != ""
        stmp = Builtins.splitstring(passwdData, "\n") # get individual lines
        nameTmp = []

        Builtins.foreach(stmp) do |user|
          next if user == nil || user == ""
          nameTmp = Builtins.splitstring(user, ":") # get individual elements
          found = false
          Builtins.foreach(lumUsers) do |name|
            if Ops.get(nameTmp, 0, "") == name
              Builtins.y2milestone(
                "nss.DeleteLumUsers:found a user to delete:%1",
                name
              )
              found = true
              raise Break
            end
          end
          if found
            next
          else
            buildStr = Ops.add(Ops.add(buildStr, user), "\n")
          end
        end
        if buildStr != ""
          if SCR.Write(path(".target.string"), passwdFile, buildStr) != true
            Builtins.y2milestone(
              "nss.DeleteLumUsers:Could not write %1",
              passwdFile
            )
          else
            bRet = true
          end
        end
      end
      bRet
    end

    def DeleteLumGroups(lumGroups)
      lumGroups = deep_copy(lumGroups)
      bRet = false

      groupFile = "/etc/group"

      # Backup the /etc/group file
      SCR.Execute(path(".target.bash"), "cp /etc/group /etc/group_OESsave")

      # Read the file into memory
      groupData = Convert.to_string(SCR.Read(path(".target.string"), groupFile))
      buildStr = ""

      if groupData != nil && groupData != ""
        stmp = Builtins.splitstring(groupData, "\n") # get individual lines
        nameTmp = []

        Builtins.foreach(stmp) do |group|
          next if group == nil || group == ""
          nameTmp = Builtins.splitstring(group, ":") # get individual elements
          found = false
          Builtins.foreach(lumGroups) do |name|
            if Ops.get(nameTmp, 0, "") == name
              Builtins.y2milestone(
                "nss.DeleteLumGroups:found a group to delete:%1",
                name
              )
              found = true
              raise Break
            end
          end
          if found
            next
          else
            buildStr = Ops.add(Ops.add(buildStr, group), "\n")
          end
        end
        if buildStr != ""
          if SCR.Write(path(".target.string"), groupFile, buildStr) != true
            Builtins.y2milestone(
              "nss.DeleteLumGroups:Could not write %1",
              groupFile
            )
          else
            bRet = true
          end
        end
      end
      bRet
    end

    def EdirRetryDone
      ret = false

      if Ops.greater_than(
          SCR.Read(
            path(".target.size"),
            "/opt/novell/oes-install/edirretryfailurer"
          ),
          0
        )
        ret = true
        SCR.Execute(
          path(".target.bash"),
          "rm -f /opt/novell/oes-install/edirretryfailure"
        ) # Get rid edir failure flag
      end
      Builtins.y2milestone("Nss.EdirRetryDone:ret: %1", ret)
      ret
    end

    # This method probably not needed anymore
    #global string RecalculateNssAdminDN()
    #{
    #    string wrkContext = "";
    #    if( Mode::installation() && Stage::normal() && edir_service_configured )
    #    {
    #        wrkContext = Misc::SysconfigRead(.sysconfig.edirectory.CONFIG_EDIR_SERVER_CONTEXT,"");
    #	y2milestone("Nss:RecalculateNssAdminDN (post install mode).eDir server context from conf file: %1", wrkContext);
    #    }
    #    else
    #    {
    #        wrkContext = Edirectory::EdirSettings["CONFIG_EDIR_SERVER_CONTEXT"]:"";
    #        y2milestone("Nss:RecalculateNssAdminDN.eDir server context from map: %1", wrkContext);
    #        if( wrkContext == nil || wrkContext == "" )
    #        {
    #            wrkContext = Misc::SysconfigRead(.sysconfig.edirectory.CONFIG_EDIR_SERVER_CONTEXT,"");
    #	    y2milestone("Nss:RecalculateNssAdminDN.eDir server context from conf file: %1", wrkContext);
    #        }
    #    }

    #    if( wrkContext == nil || wrkContext == "" )
    #    {
    #        wrkContext = NovellUtils::GetContextNDS(NovellLDAP::Settings["CONFIG_LDAP_ADMIN_CONTEXT"]:"");
    #        y2milestone("Nss:RecalculateNssAdminDN.admin user context (could not get eDir server context): %1", wrkContext);
    #        if( wrkContext == nil || wrkContext == "" ) wrkContext = "";
    #    }

    #    if( xadMode )
    #    {
    #        string nameMap = GetNameMap();
    #        string nameMapContext = GetNameMapContainer();

    #        if( nameMap == "yes" )
    #            wrkContext = oesContext + "." + NovellUtils::Ldap2Nds( nameMapContext );
    #	else
    #            wrkContext = NovellUtils::Ldap2Nds( wrkContext );
    #    }
    #    Settings["CONFIG_NSS_NSSADMIN_DN"] = "cn=" + NovellUtils::GetHostname() + "admin." + NovellUtils::Ldap2Nds( wrkContext );

    #    y2milestone("Nss:RecalculateNssAdminDN: returning %1", Settings["CONFIG_NSS_NSSADMIN_DN"]:"");
    #    return Settings["CONFIG_NSS_NSSADMIN_DN"]:"";
    #}

    # Write all nss settings
    # @return true on success
    def Write
      Builtins.y2milestone("Nss::Write started")
      Wizard.DisableAbortButton
      pattern = "novell-nss"
      if NovellPatterns.IsPatternInstalled(pattern) == false
        Builtins.y2milestone(
          "The required pattern, %1, is not installed.",
          pattern
        )
        if Mode.normal
          Report.Error(
            Builtins.sformat(
              _(
                "Configuration aborted. Install the required %1 pattern and try again."
              ),
              pattern
            )
          )
        end

        return true
      end
      if !Ops.get(NovellAddonInstall.products_to_configure, "nss", true)
        Builtins.y2milestone("NSS -- configure later")
        return true
      end

      abortInstall = Builtins.tolower(
        NovellUtils.ReadOption("ABORT_OES_INSTALL")
      )
      if abortInstall == "yes"
        Builtins.y2milestone("Nss:setting ABORT_OES_INSTALL is set to 'yes'.")
        return true
      end
      # Nss read dialog caption
      caption = _("Saving OES Storage Services Configuration")

      # TODO FIXME And set the right number of stages
      steps = 2

      sl = 100
      Builtins.sleep(sl)

      # TODO FIXME Names of real stages
      # We do not set help text here, because it was set outside
      Progress.New(
        caption,
        " ",
        steps,
        [
          # Progress stage 1/1
          _("Write the OES Storage Services settings")
        ],
        [
          # Progress step 1/1
          _("Writing the settings..."),
          # Progress finished
          _("Finished")
        ],
        ""
      )


      if Service.Status("namcd") != 0
        Builtins.y2milestone("namcd service is down, making it up")
        if Service.Start("namcd")
          Builtins.y2milestone("namcd service started successfully")
        else
          Builtins.y2milestone("failed to start namcd service")
          Report.Error(
            Builtins.sformat(
              _("NSS configuration failed. Ensure namcd service is running.")
            )
          )
          return true
        end
      end

      # If XAD did not install properly, then we don't want to proceed at this point.
      @xad_packages_installed = NovellPatterns.IsPatternInstalled(
        "novell-domain-services"
      )
      Builtins.y2milestone(
        "Nss:xad_packages_installed:%1",
        @xad_packages_installed
      )
      if @xad_packages_installed
        if Ops.get_string(Edirectory.EdirSettings, "XAD_SERVICE_CONFIGURED", "") == "yes" ||
            Convert.to_string(
              SCR.Read(path(".sysconfig.novell.xad.SERVICE_CONFIGURED"))
            ) == "yes"
          @xadMode = true
        end
      end

      @xadMode = true if @xadTestMode

      Builtins.y2milestone("Nss:xadMode:%1", @xadMode)

      Builtins.y2milestone(
        "Nss:testing to see if eDir configured properly. Xad mode: %1",
        @xadMode
      )
      if @xadMode == true
        Builtins.y2milestone(
          "Nss:eDir service_configured (mem):%1",
          Ops.get_string(Edirectory.EdirSettings, "SERVICE_CONFIGURED", "")
        )
        Builtins.y2milestone(
          "Nss:eDir service_configured (file):%1",
          Convert.to_string(
            SCR.Read(path(".sysconfig.edirectory.SERVICE_CONFIGURED"))
          )
        )
        if Ops.get_string(Edirectory.EdirSettings, "SERVICE_CONFIGURED", "") != "yes" &&
            Convert.to_string(
              SCR.Read(path(".sysconfig.edirectory.SERVICE_CONFIGURED"))
            ) != "yes"
          if EdirUtils.EdirConfigProblemPrompt("NSS")
            Builtins.y2milestone(
              "Nss:aborting NSS write by user choice since eDirectory was not configured."
            )
            return true
          end
        end
      end


      reinstallStr = ""
      reinstallStr = "REINSTALL='1' " if @reinstallMode

      # Since eDir may have failed during Write() and the user may have changed the server context to fix that issue, we will recalculate values based on current eDir context.
      # The original eDir context obtained during Read would no longer be valid (#639758).
      if @userChangedData == false
        if @doRecalculate # if we are in autoYaST mode, upgrade mode, or reconfigure mode, then we don't want to recalculate
          CalculateAndSetNssAdminAndServerContext()
        end
      end

      # determine IP addresses
      server_list = Ops.get_list(
        NovellLDAP.Settings,
        "CONFIG_LDAP_SERVERS_LIST",
        []
      )
      primary_server = NovellLDAP.GetAddressServer(
        Ops.get_string(@Settings, "CONFIG_NSS_LDAP_SERVER", ""),
        server_list
      )
      fallback_list = NovellLDAP.GetFallbackServerList(
        Ops.get_string(@Settings, "CONFIG_NSS_LDAP_SERVER", ""),
        server_list
      )
      Builtins.y2milestone("Primary server: %1", primary_server)
      Builtins.y2milestone("Fallback server_list: %1", fallback_list)

      # write settings
      return false if Abort()
      Progress.NextStage

      SCR.Write(path(".sysconfig.nss.SERVICE_CONFIGURED"), "no")
      SCR.Write(
        path(".sysconfig.nss.CONFIG_NSS_LDAP_SERVER"),
        Ops.get_string(@Settings, "CONFIG_NSS_LDAP_SERVER", "")
      )
      SCR.Write(
        path(".sysconfig.nss.CONFIG_NSS_NSSADMIN_DN"),
        Ops.get_string(@Settings, "CONFIG_NSS_NSSADMIN_DN", "")
      )

      # Flush the sysconfig file. NSS must be able to read these values from the file before the calls below.
      SCR.Write(path(".sysconfig.nss"), nil)

      if !Mode.test
        # Don't delete the local users and groups if LUM enablement fails.
        deleteLocalsOk = true
        deleteXtierLocalsOk = true

        nssldconffile = "/etc/ld.so.conf.d/novell-nss.conf"
        if SCR.Read(path(".target.size"), nssldconffile) == -1
          ret2 = Convert.to_integer(
            SCR.Execute(
              path(".target.bash"),
              Ops.add("echo /opt/novell/lib > ", nssldconffile)
            )
          )
          ret2 = Convert.to_integer(
            SCR.Execute(path(".target.bash"), "ldconfig")
          )
        end

        # Get Settings Variables
        fqhostname = NovellUtils.GetIPAddr
        hostname = NovellUtils.GetHostname

        admin_dn = NovellUtils.Ldap2Nds(
          Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_ADMIN_CONTEXT", "")
        )
        admin_password = Ops.get_string(
          NovellLDAP.Settings,
          "CONFIG_LDAP_ADMIN_PASSWORD",
          ""
        )

        if @xadMode
          xad_tree_admin_context = Ops.get_string(
            NovellLDAP.Settings,
            "XAD_TREE_ADMIN_CONTEXT",
            ""
          )
          Builtins.y2milestone(
            "Nss:XAD_TREE_ADMIN_CONTEXT:%1",
            xad_tree_admin_context
          )
          if xad_tree_admin_context != nil && xad_tree_admin_context != ""
            xad_tree_admin_context = NovellUtils.Ldap2Nds(
              xad_tree_admin_context
            )
            admin_dn = xad_tree_admin_context
            admin_password = Ops.get_string(
              NovellLDAP.Settings,
              "XAD_TREE_ADMIN_PASSWORD",
              ""
            )
            Builtins.y2milestone("Nss:admin:%1", admin_dn)
          end
        end

        # Since the LUM enabling of Admin has been moved to LUM, we need to make sure we are using the same context.
        server_context = Convert.to_string(
          SCR.Read(path(".sysconfig.novell-lum.CONFIG_LUM_WS_CONTEXT"))
        )
        Builtins.y2milestone(
          "Nss.Write:LUM WS context from sysconfig read is '%1'",
          server_context
        )
        if server_context == nil || server_context == ""
          server_context = Ops.get_string(
            NovellLum.Settings,
            "CONFIG_LUM_WS_CONTEXT",
            ""
          )
          Builtins.y2milestone(
            "Nss.Write:LUM WS context from map value is '%1'",
            server_context
          )
        end

        #        string NSSAdminDN = "";
        #        if( EdirRetryDone() == false )
        #	{


        _NSSAdminDN = NovellUtils.Ldap2Nds(
          Ops.get_string(@Settings, "CONFIG_NSS_NSSADMIN_DN", "")
        )



        #	}
        #	else
        #	{
        #            NSSAdminDN = RecalculateNssAdminDN();  // probably no longer needed
        #            Settings["CONFIG_NSS_NSSADMIN_DN"] = NSSAdminDN;
        #            SCR::Write(.sysconfig.nss.CONFIG_NSS_NSSADMIN_DN, Settings["CONFIG_NSS_NSSADMIN_DN"]:"");
        #	}

        ip_or_hostname = fqhostname

        upgradeEnvVar = ""
        if Mode.update || Mode.autoupgrade
          upgradeEnvVar = "UPGRADE=1 "
        end

        # Start ID Broker

        useServerContext = false
        serverContext = Ops.get_string(
          @Settings,
          "CONFIG_NSS_EDIR_SERVER_CONTEXT",
          ""
        )
        nssAdminContext = NovellUtils.GetContextNDS(_NSSAdminDN)
        serverContext = NovellUtils.Ldap2Nds(serverContext)
        Builtins.y2milestone(
          "Nss.Write:eDir server context is %1",
          serverContext
        )
        nssAdminContext = NovellUtils.Ldap2Nds(nssAdminContext)
        Builtins.y2milestone(
          "Nss.Write:nssAdmin context is %1",
          nssAdminContext
        )

        if Builtins.tolower(serverContext) != Builtins.tolower(nssAdminContext)
          useServerContext = true
          Builtins.y2milestone(
            "Nss.Write:eDir server context and nss admin context were different, so we will use the -c option."
          )
        else
          Builtins.y2milestone(
            "Nss.Write:eDir server context and nss admin context were the same, so we will NOT use the -c option."
          )
        end

        nssadmin_install_file = "/opt/novell/nss/sbin/nssAdminInstall"

        # nssAdminInstall -a 'admin.novell' -p n0v3ll -o 'nssadminuser.blr.in' -c novell

        execute_cmd = Ops.add(
          Ops.add(
            Ops.add(
              Ops.add(Ops.add(nssadmin_install_file, " -a '"), admin_dn),
              "' -o '"
            ),
            _NSSAdminDN
          ),
          "'"
        )
        if useServerContext
          execute_cmd = Ops.add(Ops.add(execute_cmd, " -c "), serverContext)
        end

        Builtins.y2milestone("NSS:Execute %1", execute_cmd)
        ret = Convert.to_integer(
          SCR.Execute(
            path(".target.bash"),
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(Ops.add(upgradeEnvVar, reinstallStr), "OES_DATA='"),
                  admin_password
                ),
                "' "
              ),
              execute_cmd
            )
          )
        )
        if ret != 0
          Report.Error(
            _("NSS Admin install failed creating eDirectory objects.")
          )
        end

        # Extend the eDirectory schema (.sch files)
        ndssch_file = "/opt/novell/eDirectory/bin/ndssch"
        schema_dir = "/opt/novell/nss/sbin/"
        schema_files = ["ndfs1.sch", "nssfs.sch","cis.sch"]
        fallback_ip = Ops.get_string(@Settings, "CONFIG_FALLBACK_IP", "")
        Builtins.foreach(schema_files) do |schema_file|
          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(ndssch_file, " -h "),
                          Ops.get_string(
                            primary_server,
                            "CONFIG_LDAP_SERVER_IP",
                            ""
                          )
                        ),
                        " -F '"
                      ),
                      @schemaLogPath
                    ),
                    "' '"
                  ),
                  admin_dn
                ),
                "' "
              ),
              schema_dir
            ),
            schema_file
          )
          Builtins.y2milestone("Nss:executing:%1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "xad:Nss:executing:%1",
              Ops.add(
                Ops.add(Ops.add("ADM_PASSWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          ret = Convert.to_integer(
            SCR.Execute(
              path(".target.bash"),
              Ops.add(
                Ops.add(Ops.add("ADM_PASSWD='", admin_password), "' "),
                execute_cmd
              )
            )
          )
          if ret != 0
            Builtins.y2milestone(
              "Nss:Failed (error=%1) to extend schema using default IP address / hostname: %2, using file %3",
              ret,
              Ops.get_string(primary_server, "CONFIG_LDAP_SERVER_IP", ""),
              schema_file
            )
            # Try the fallback IP if it exists.
            Builtins.foreach(fallback_list) do |server_info|
              execute_cmd = Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(ndssch_file, " -h "),
                              Ops.get_string(
                                server_info,
                                "CONFIG_LDAP_SERVER_IP",
                                ""
                              )
                            ),
                            " -F '"
                          ),
                          @schemaLogPath
                        ),
                        "' '"
                      ),
                      admin_dn
                    ),
                    "' "
                  ),
                  schema_dir
                ),
                schema_file
              )
              Builtins.y2milestone("Nss:executing:%1", execute_cmd)
              if @xadTestMode
                Builtins.y2milestone(
                  "xac:Nss:executing:%1",
                  Ops.add(
                    Ops.add(Ops.add("ADM_PASSWD='", admin_password), "' "),
                    execute_cmd
                  )
                )
              end
              ret = Convert.to_integer(
                SCR.Execute(
                  path(".target.bash"),
                  Ops.add(
                    Ops.add(Ops.add("ADM_PASSWD='", admin_password), "' "),
                    execute_cmd
                  )
                )
              )
              if ret != 0
                Builtins.y2milestone(
                  "Nss:Failed (error=%1) to extend schema using fallback IP address / hostname: %2, using file %3",
                  ret,
                  fallback_ip,
                  schema_file
                )
              else
                raise Break
              end
            end
          end
          if ret != 0
            Builtins.y2milestone(
              "NSS install failed extending schema file %1",
              schema_file
            )
          end
        end 


        #**************************************************************************************************
        #
        # We must also add the group www and user wwwrun to eDir so they can access NSS volumes.
        # We do this by doing the following:
        # We use the eDirectory server context that was assigned in eDir. We first try a namgroup
        # add of the group www. If this succeeds, we know we are
        # adding this group for the first time. If namgroupadd fails, but namgroupmodify succeeds, then we know
        # we are adding an additional server into the mix.  Depending on these states, different things are done.
        #
        # For the "first add" group state, we do the following:
        #   1) namgroupadd www.
        #   2) namuseradd wwwrun
        #   3) Delete local version of wwwrun
        #
        # If we are adding a new server, we do the following:
        #   1) Get the GID of group www from LUM
        #   2) namgroupmod www
        #   3) Get the UID of user wwwrun from LUM.
        #   4) Delete local version of wwwrun
        #  (mtc)
        #**************************************************************************************************

        Builtins.y2milestone(
          "Nss:adding group www and user wwwrun to eDir so they can access NSS volumes."
        )

        # First, shut down the nscd daemon so we can delete our local groups.
        execute_cmd = "systemctl stop nscd.service"
        Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
        rmap = Convert.to_map(
          SCR.Execute(path(".target.bash_output"), execute_cmd)
        )
        if Ops.get_integer(rmap, "exit", -1) != 0
          Builtins.y2milestone(
            "Nss:Error on namgroupadd for group www (stdout): %1",
            Ops.get_string(rmap, "stdout", "")
          )
          Builtins.y2milestone(
            "Nss:Error on namgroupadd for group www (stderr): %1",
            Ops.get_string(rmap, "stderr", "")
          )
        end

        l_www_gid = NovellUtils.Get_GID("www")
        l_wwwrun_uid = NovellUtils.Get_UID("wwwrun")

        local_www_gid = Ops.add(NovellUtils.Get_GID("www"), " -o -f")
        local_wwwrun_uid = Ops.add(NovellUtils.Get_UID("wwwrun"), " -o -f")

        Builtins.y2milestone("Nss:www GID: %1", local_www_gid)
        Builtins.y2milestone("Nss:wwwrun UID: %1", local_wwwrun_uid)

        all_worked = true

        server_context = NovellUtils.Nds2Ldap(server_context)
        server_context = "" if server_context == nil # avoid nil value
        Builtins.sleep(5000)
        execute_cmd = Ops.add(
          Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        "namgroupadd -a '",
                        NovellUtils.Nds2Ldap(admin_dn)
                      ),
                      "' -x '"
                    ),
                    server_context
                  ),
                  "' -W "
                ),
                hostname
              ),
              " -g "
            ),
            local_www_gid
          ),
          " www"
        )
        Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
        if @xadTestMode
          Builtins.y2milestone(
            "xad:Nss:execute command: %1",
            Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
          )
        end
        execute_cmd = Ops.add(
          Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
          execute_cmd
        )
        rmap = Convert.to_map(
          SCR.Execute(path(".target.bash_output"), execute_cmd)
        )

        if Ops.get_integer(rmap, "exit", -1) != 0 # We are adding another server to the mix.
          Builtins.y2milestone(
            "Nss:Error on namgroupadd for group www (stdout): %1",
            Ops.get_string(rmap, "stdout", "")
          )
          Builtins.y2milestone(
            "Nss:Error on namgroupadd for group www (stderr): %1",
            Ops.get_string(rmap, "stderr", "")
          )
          Builtins.y2milestone(
            "Nss:Since the namgroupadd failed, we will assume we are adding an additional server to the mix."
          )
          # The above call failed, so the group could already be there, in which case we must do a modify rather than an add.
          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add("namgroupmod -a '", NovellUtils.Nds2Ldap(admin_dn)),
                    "' -W "
                  ),
                  hostname
                ),
                " 'cn=www,"
              ),
              server_context
            ),
            "'"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "xad:Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namgroupmod for group www (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namgroupmod for group www (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            all_worked = false
            deleteLocalsOk = false
          end
          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  "namuseradd -a '",
                                  NovellUtils.Nds2Ldap(admin_dn)
                                ),
                                "' -x '"
                              ),
                              server_context
                            ),
                            "'"
                          ),
                          " -d /var/lib/wwwrun"
                        ),
                        " -g "
                      ),
                      "'cn=www,"
                    ),
                    server_context
                  ),
                  "' -u "
                ),
                local_wwwrun_uid
              ),
              " -s /bin/false "
            ),
            "wwwrun"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error (may not indicate a real error) on namuseradd for user wwwrun (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user wwwrun (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  "namusermod -a '",
                                  NovellUtils.Nds2Ldap(admin_dn)
                                ),
                                "'"
                              ),
                              " -d /var/lib/wwwrun"
                            ),
                            " -g "
                          ),
                          "'cn=www,"
                        ),
                        server_context
                      ),
                      "'"
                    ),
                    " -s /bin/false "
                  ),
                  "'cn=wwwrun,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end

            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteLocalsOk = false
            end
          end

          if all_worked == true
            lum_wwwGID = GetGIDfromLUM("www")
            lum_wwwrunUID = GetUIDfromLUM("wwwrun")

            Builtins.y2milestone("Nss:lum www GID: %1", lum_wwwGID)
            Builtins.y2milestone("Nss:lum wwwrun UID: %1", lum_wwwrunUID)

            # Instead of doing this in the install, which takes hours, we will write out a script file that the user will execute at his leasure.
            if !Mode.update && !Mode.autoupgrade
              retbool = WriteScriptFileForChown(
                l_www_gid,
                lum_wwwGID,
                l_wwwrun_uid,
                lum_wwwrunUID
              )
              retbool = ChangeModForNssidFile()

              warn = _(
                "The script file /opt/novell/oes-install/nssid.sh must be manually run\n" +
                  "to complete the configuration of NSS.\n" +
                  "Because scanning the file system may take some time,\n" +
                  "the system administrator should run this script after the install completes.\n" +
                  "NSS will not be completely configured until the script is run."
              )
              Builtins.y2milestone(warn)
              # removed at the request of management		Report::Warning(warn);
            end
          end # First time group and users added to eDir.
        else
          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  "namuseradd -a '",
                                  NovellUtils.Nds2Ldap(admin_dn)
                                ),
                                "' -x '"
                              ),
                              server_context
                            ),
                            "'"
                          ),
                          " -d /var/lib/wwwrun"
                        ),
                        " -g "
                      ),
                      "'cn=www,"
                    ),
                    server_context
                  ),
                  "' -u "
                ),
                local_wwwrun_uid
              ),
              " -s /bin/false "
            ),
            "wwwrun"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error (may not indicate a real error) on namuseradd for user wwwrun (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user wwwrun (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  "namusermod -a '",
                                  NovellUtils.Nds2Ldap(admin_dn)
                                ),
                                "'"
                              ),
                              " -d /var/lib/wwwrun"
                            ),
                            " -g "
                          ),
                          "'cn=www,"
                        ),
                        server_context
                      ),
                      "'"
                    ),
                    " -s /bin/false "
                  ),
                  "'cn=wwwrun,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteLocalsOk = false
            end
          end
        end

        #**************************************************************************************************
        #
        # We must also add the group novlxtier, and users novlxregd and novlxsrvd to eDir.
        # We do this by doing the following:
        # We use the eDir server context that was assigned in eDir. We first try a namgroupadd of the
        # group novlxtier. If this succeeds, we know we are
        # adding this group for the first time. If namgroupadd fails, but namgroupmodify succeeds, then we know
        # we are adding an additional server into the mix.  Depending on these states, different things are done.
        #
        # For the "first add" group state, we do the following:
        #   1) namgroupadd novlxtier.
        #   2) namuseradd novlxregd
        #   3) namuseradd novlxsrvd
        #   4) Delete local versions of novlxregd and novlxsrvd if they exist
        #
        # If we are adding a new server, we do the following:
        #   1) Get the GID of group novlxtier from LUM
        #   2) namgroupmod novlxtier
        #   3) Get the UIDs of users novlxregs and novlxsrvd from LUM.
        #   4) Delete local versions of novlxregs and novlxsrvd if they exist.
        #   5) Change ownership of all files in the filesystem with novellxregd:novlxtier to match new GID and UID.
        #   6) Change ownership of all files in the filesystem with novlxsrvd:novlxtier to match new GID and UID.
        #
        #  Warning: steps 5 and 6 take a LONG LONG time - this has since been moved to a script file that the user executes.
        #  (mtc)
        #**************************************************************************************************

        Builtins.y2milestone(
          "Nss:adding group novlxtier and users novlxregd and novlxsrvd to eDir."
        )

        # If there are no local xtier users or groups, we will create them as an easy way to get system GIDs and UIDs.
        local_novlxtier_gid = GetOrCreateGID("novlxtier")
        local_novlxregd_uid = GetOrCreateUID("novlxregd", "novlxtier")
        local_novlxsrvd_uid = GetOrCreateUID("novlxsrvd", "novlxtier")

        Builtins.y2milestone("Nss:novlxtier GID: %1", local_novlxtier_gid)
        Builtins.y2milestone("Nss:novlxregd UID: %1", local_novlxregd_uid)
        Builtins.y2milestone("Nss:novlxsrvd UID: %1", local_novlxsrvd_uid)

        all_worked = true

        gidSection = " "
        if local_novlxtier_gid != ""
          gidSection = Ops.add(Ops.add(" -g ", local_novlxtier_gid), " -o -f ")
        end

        Builtins.sleep(5000)
        execute_cmd = Ops.add(
          Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add("namgroupadd -a '", NovellUtils.Nds2Ldap(admin_dn)),
                    "' -x '"
                  ),
                  server_context
                ),
                "' -W "
              ),
              hostname
            ),
            gidSection
          ),
          "novlxtier"
        )
        Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
        if @xadTestMode
          Builtins.y2milestone(
            "Nss:execute command: %1",
            Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
          )
        end
        execute_cmd = Ops.add(
          Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
          execute_cmd
        )
        rmap = Convert.to_map(
          SCR.Execute(path(".target.bash_output"), execute_cmd)
        )

        if Ops.get_integer(rmap, "exit", -1) != 0 # We are adding another server to the mix.
          Builtins.y2milestone(
            "Nss:Error on namgroupadd for group novlxtier (stdout): %1",
            Ops.get_string(rmap, "stdout", "")
          )
          Builtins.y2milestone(
            "Nss:Error on namgroupadd for group novlxtier (stderr): %1",
            Ops.get_string(rmap, "stderr", "")
          )
          Builtins.y2milestone(
            "Nss:Since the namgroupadd failed, we will assume we are adding an additional server to the mix."
          )

          # The above call failed, so the group could already be there, in which case we must do a modify rather than an add.
          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add("namgroupmod -a '", NovellUtils.Nds2Ldap(admin_dn)),
                    "' -W "
                  ),
                  hostname
                ),
                " 'cn=novlxtier,"
              ),
              server_context
            ),
            "'"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namgroupmod for group novlxtier (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namgroupmod for group novlxtier (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            all_worked = false
            deleteXtierLocalsOk = false
          end

          uidSectionXregd = " "
          if local_novlxregd_uid != ""
            uidSectionXregd = Ops.add(
              Ops.add(" -u ", local_novlxregd_uid),
              " -o -f "
            )
          end

          uidSectionXsrvd = " "
          if local_novlxsrvd_uid != ""
            uidSectionXsrvd = Ops.add(
              Ops.add(" -u ", local_novlxsrvd_uid),
              " -o -f "
            )
          end

          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                "namuseradd -a '",
                                NovellUtils.Nds2Ldap(admin_dn)
                              ),
                              "' -x '"
                            ),
                            server_context
                          ),
                          "' -c 'Novell XRegD System User'"
                        ),
                        " -d /var/opt/novell/xtier/xregd"
                      ),
                      " -g "
                    ),
                    "'cn=novlxtier,"
                  ),
                  server_context
                ),
                "' -s /bin/false"
              ),
              uidSectionXregd
            ),
            "novlxregd"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxregd (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxregd (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )

            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  Ops.add(
                                    "namusermod -a '",
                                    NovellUtils.Nds2Ldap(admin_dn)
                                  ),
                                  "'"
                                ),
                                " -c 'Novell XRegD System User'"
                              ),
                              " -d /var/opt/novell/xtier/xregd"
                            ),
                            " -g "
                          ),
                          "'cn=novlxtier,"
                        ),
                        server_context
                      ),
                      "' -s /bin/false"
                    ),
                    " "
                  ),
                  "'cn=novlxregd,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxregd (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxregd (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end

          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                "namuseradd -a '",
                                NovellUtils.Nds2Ldap(admin_dn)
                              ),
                              "' -x '"
                            ),
                            server_context
                          ),
                          "' -c 'Novell XSrvD System User'"
                        ),
                        " -d /var/opt/novell/xtier/xsrvd"
                      ),
                      " -g "
                    ),
                    "'cn=novlxtier,"
                  ),
                  server_context
                ),
                "' -s /bin/false"
              ),
              uidSectionXsrvd
            ),
            "novlxsrvd"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxsrvd (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxsrvd (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  Ops.add(
                                    "namusermod -a '",
                                    NovellUtils.Nds2Ldap(admin_dn)
                                  ),
                                  "'"
                                ),
                                " -c 'Novell XSrvD System User'"
                              ),
                              " -d /var/opt/novell/xtier/xsrvd"
                            ),
                            " -g "
                          ),
                          "'cn=novlxtier,"
                        ),
                        server_context
                      ),
                      "' -s /bin/false"
                    ),
                    " "
                  ),
                  "'cn=novlxsrvd,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end

          # Make the required associations:
          #   wwwrun needs to be a member of the novlxtier group.
          #   novlxsrvd need to be a member of the www group.
          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            "namusermod -a '",
                            NovellUtils.Nds2Ldap(admin_dn)
                          ),
                          "' -g "
                        ),
                        "'cn=www,"
                      ),
                      server_context
                    ),
                    "' -G 'cn=novlxtier,"
                  ),
                  server_context
                ),
                "' 'cn=wwwrun,"
              ),
              server_context
            ),
            "'"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namusermod for user wwwrun - retrying without -g option (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namusermod for user wwwrun (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          "namusermod -a '",
                          NovellUtils.Nds2Ldap(admin_dn)
                        ),
                        "'"
                      ),
                      " -G 'cn=novlxtier,"
                    ),
                    server_context
                  ),
                  "' 'cn=wwwrun,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end

          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            "namusermod -a '",
                            NovellUtils.Nds2Ldap(admin_dn)
                          ),
                          "' -g "
                        ),
                        "'cn=novlxtier,"
                      ),
                      server_context
                    ),
                    "' -G 'cn=www,"
                  ),
                  server_context
                ),
                "' 'cn=novlxsrvd,"
              ),
              server_context
            ),
            "'"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namusermod for user novlxsrvd - retrying without the -g option (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namusermod for user novlxsrvd (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          "namusermod -a '",
                          NovellUtils.Nds2Ldap(admin_dn)
                        ),
                        "'"
                      ),
                      " -G 'cn=www,"
                    ),
                    server_context
                  ),
                  "' 'cn=novlxsrvd,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end


          if all_worked == true
            lum_novlxtierGID = GetGIDfromLUM("novlxtier")
            lum_novlxregdUID = GetUIDfromLUM("novlxregd")
            lum_novlxsrvdUID = GetUIDfromLUM("novlxsrvd")

            Builtins.y2milestone("Nss:lum novlxtier GID: %1", lum_novlxtierGID)
            Builtins.y2milestone("Nss:lum novlxregd UID: %1", lum_novlxregdUID)
            Builtins.y2milestone("Nss:lum novlxsrvd UID: %1", lum_novlxsrvdUID)

            # We will write out a script file that the user will execute at his leasure - perhaps in the background overnight..
            if !Mode.update && !Mode.autoupgrade
              AppendScriptFileForXtierChown(
                local_novlxregd_uid,
                local_novlxtier_gid,
                lum_novlxregdUID,
                lum_novlxtierGID,
                local_novlxsrvd_uid,
                lum_novlxsrvdUID
              )
              retbool = ChangeModForNssidFile()
            end
          end # First time group and users added to eDir.
        else
          uidSectionXregd = " "
          if local_novlxregd_uid != ""
            uidSectionXregd = Ops.add(
              Ops.add(" -u ", local_novlxregd_uid),
              " -o -f "
            )
          end

          uidSectionXsrvd = " "
          if local_novlxsrvd_uid != ""
            uidSectionXsrvd = Ops.add(
              Ops.add(" -u ", local_novlxsrvd_uid),
              " -o -f "
            )
          end

          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                "namuseradd -a '",
                                NovellUtils.Nds2Ldap(admin_dn)
                              ),
                              "' -x '"
                            ),
                            server_context
                          ),
                          "' -c 'Novell XRegD System User'"
                        ),
                        " -d /var/opt/novell/xtier/xregd"
                      ),
                      " -g "
                    ),
                    "'cn=novlxtier,"
                  ),
                  server_context
                ),
                "' -s /bin/false"
              ),
              uidSectionXregd
            ),
            "novlxregd"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxregd (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxregd (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )

            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  Ops.add(
                                    "namusermod -a '",
                                    NovellUtils.Nds2Ldap(admin_dn)
                                  ),
                                  "'"
                                ),
                                " -c 'Novell XRegD System User'"
                              ),
                              " -d /var/opt/novell/xtier/xregd"
                            ),
                            " -g "
                          ),
                          "'cn=novlxtier,"
                        ),
                        server_context
                      ),
                      "' -s /bin/false"
                    ),
                    " "
                  ),
                  "'cn=novlxregd,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxregd (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxregd (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end

          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                "namuseradd -a '",
                                NovellUtils.Nds2Ldap(admin_dn)
                              ),
                              "' -x '"
                            ),
                            server_context
                          ),
                          "' -c 'Novell XSrvD System User'"
                        ),
                        " -d /var/opt/novell/xtier/xsrvd"
                      ),
                      " -g "
                    ),
                    "'cn=novlxtier,"
                  ),
                  server_context
                ),
                "' -s /bin/false"
              ),
              uidSectionXsrvd
            ),
            "novlxsrvd"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxsrvd (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namuseradd for user novlxsrvd (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(
                              Ops.add(
                                Ops.add(
                                  Ops.add(
                                    "namusermod -a '",
                                    NovellUtils.Nds2Ldap(admin_dn)
                                  ),
                                  "'"
                                ),
                                " -c 'Novell XSrvD System User'"
                              ),
                              " -d /var/opt/novell/xtier/xsrvd"
                            ),
                            " -g "
                          ),
                          "'cn=novlxtier,"
                        ),
                        server_context
                      ),
                      "' -s /bin/false"
                    ),
                    " "
                  ),
                  "'cn=novlxsrvd,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end

          # Make the required associations:
          #   wwwrun needs to be a member of the novlxtier group.
          #   novlxsrvd need to be a member of the www group.
          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            "namusermod -a '",
                            NovellUtils.Nds2Ldap(admin_dn)
                          ),
                          "' -g "
                        ),
                        "'cn=www,"
                      ),
                      server_context
                    ),
                    "' -G 'cn=novlxtier,"
                  ),
                  server_context
                ),
                "' 'cn=wwwrun,"
              ),
              server_context
            ),
            "'"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namusermod for user wwwrun - retrying without -g option (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namusermod for user wwwrun (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          "namusermod -a '",
                          NovellUtils.Nds2Ldap(admin_dn)
                        ),
                        "'"
                      ),
                      " -G 'cn=novlxtier,"
                    ),
                    server_context
                  ),
                  "' 'cn=wwwrun,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user wwwrun (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end

          execute_cmd = Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            "namusermod -a '",
                            NovellUtils.Nds2Ldap(admin_dn)
                          ),
                          "' -g "
                        ),
                        "'cn=novlxtier,"
                      ),
                      server_context
                    ),
                    "' -G 'cn=www,"
                  ),
                  server_context
                ),
                "' 'cn=novlxsrvd,"
              ),
              server_context
            ),
            "'"
          )
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode
            Builtins.y2milestone(
              "Nss:execute command: %1",
              Ops.add(
                Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                execute_cmd
              )
            )
          end
          execute_cmd = Ops.add(
            Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
            execute_cmd
          )
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error on namusermod for user novlxsrvd - retrying without the -g option (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
            Builtins.y2milestone(
              "Nss:Error on namusermod for user novlxsrvd (stderr): %1",
              Ops.get_string(rmap, "stderr", "")
            )
            execute_cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          "namusermod -a '",
                          NovellUtils.Nds2Ldap(admin_dn)
                        ),
                        "'"
                      ),
                      " -G 'cn=www,"
                    ),
                    server_context
                  ),
                  "' 'cn=novlxsrvd,"
                ),
                server_context
              ),
              "'"
            )
            Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
            if @xadTestMode
              Builtins.y2milestone(
                "Nss:execute command: %1",
                Ops.add(
                  Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
                  execute_cmd
                )
              )
            end
            execute_cmd = Ops.add(
              Ops.add(Ops.add("LUM_PWD='", admin_password), "' "),
              execute_cmd
            )
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
            if Ops.get_integer(rmap, "exit", -1) != 0
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stdout): %1",
                Ops.get_string(rmap, "stdout", "")
              )
              Builtins.y2milestone(
                "Nss:Error on namusermod for user novlxsrvd (stderr): %1",
                Ops.get_string(rmap, "stderr", "")
              )
              all_worked = false
              deleteXtierLocalsOk = false
            end
          end
        end

        # Delete the local users and groups.
        retB = ShutDownNamcd()
        #	if( deleteLocalsOk == true )
        #	   DeleteLocals();
        #	else
        #	   y2milestone("Nss:not deleting local users, since some element of LUM enablement failed.");

        if deleteXtierLocalsOk == true
          DeleteXtierLocals()
        else
          Builtins.y2milestone(
            "Nss:not deleting xtier local users, since some element of LUM enablement failed."
          )
        end


        # Delete any temporary Groups or Users created for LUM in oes-update_proposal.ycp
        Builtins.y2milestone(
          "nss:entering code to delete any temporary lum users and groups."
        )

        usrDelFile = "/etc/sysconfig/usrdel"

        # First delete any temporary users.
        if SCR.Read(path(".target.size"), usrDelFile) != -1
          # Read the file into memory
          userData = Convert.to_string(
            SCR.Read(path(".target.string"), usrDelFile)
          )
          if userData != nil && userData != ""
            Builtins.y2milestone("nss:users to delete:%1", userData)
            bRet = DeleteLumUsers(Builtins.splitstring(userData, "\n"))
            # delete the usrDelFile
            SCR.Execute(path(".target.bash"), Ops.add("rm -f ", usrDelFile))
          end
        else
          Builtins.y2milestone("nss:no users to delete.")
        end

        # Next delete any temporary groups.
        grpDelFile = "/etc/sysconfig/grpdel"

        if SCR.Read(path(".target.size"), grpDelFile) != -1
          # Read the file into memory
          groupData = Convert.to_string(
            SCR.Read(path(".target.string"), grpDelFile)
          )
          if groupData != nil && groupData != ""
            Builtins.y2milestone("nss:groups to delete:%1", groupData)
            bRet = DeleteLumGroups(Builtins.splitstring(groupData, "\n"))
            # delete the grpDelFile
            SCR.Execute(path(".target.bash"), Ops.add("rm -f ", grpDelFile))
          end
        else
          Builtins.y2milestone("nss:no groups to delete.")
        end

        # Reset the nam cache.
        execute_cmd = "namconfig cache_refresh"
        Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
        if @xadTestMode == false
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
        end
        if Ops.get_integer(rmap, "exit", -1) != 0
          Builtins.y2milestone(
            "Nss:Error on namconfig cache_refresh (stdout): %1",
            Ops.get_string(rmap, "stdout", "")
          )
          Builtins.y2milestone(
            "Nss:Error on namconfig cache_refresh (stderr): %1",
            Ops.get_string(rmap, "stderr", "")
          )
        end

        # If namcd is not running at this point, then start it.
#        execute_cmd = "rcnamcd status"
        execute_cmd = "systemctl status namcd.service"
        Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
        if @xadTestMode == false
          rmap = Convert.to_map(
            SCR.Execute(path(".target.bash_output"), execute_cmd)
          )
        end
        if Ops.get_integer(rmap, "exit", -1) != 0
          Builtins.y2milestone(
            "Nss:namcd is not running, so we will start it(stdout): %1",
            Ops.get_string(rmap, "stdout", "")
          )
#          execute_cmd = "rcnamcd start"
          execute_cmd = "systemctl start namcd.service"
          Builtins.y2milestone("Nss:execute command: %1", execute_cmd)
          if @xadTestMode == false
            rmap = Convert.to_map(
              SCR.Execute(path(".target.bash_output"), execute_cmd)
            )
          end
          if Ops.get_integer(rmap, "exit", -1) != 0
            Builtins.y2milestone(
              "Nss:Error starting namcd (stdout): %1",
              Ops.get_string(rmap, "stdout", "")
            )
          end
        end

        # Apply NSS Settings
        ret = Convert.to_integer(
          SCR.Execute(path(".target.bash"), "mkdir /_admin")
        )

        # Make relevant files executable
        ret = Convert.to_integer(
          SCR.Execute(path(".target.bash"), "chmod +x /opt/novell/nss/sbin/*")
        )
        ret = Convert.to_integer(
          SCR.Execute(path(".target.bash"), "chmod +x /sbin/mkfs.nssvol")
        )
        ret = Convert.to_integer(
          SCR.Execute(path(".target.bash"), "chmod +x /sbin/mkfs.nsspool")
        )

        status = Service.Start("nscd")
        Builtins.y2milestone("Nss:Started nscd service : %1",status)

        #Start NSS
        Service.Enable("novell-nss")
        status = Service.Start("novell-nss")
        Builtins.y2milestone("Nss:Enabled and Started novell-nss service : %1",status)

#        #Start ncp2nss
#        Service.Enable("ncp2nss")
#        Builtins.y2milestone("Nss:Enabled ncp2nss service")

        services=["ncp2nss","novell-dfs","novell-vigil"]
        if Mode.update || Mode.autoupgrade
          Builtins.foreach(services) do |service|
            NovellUtils.SystemdStartorRestartService(service,"yes")
          end
        else
          services=["ncp2nss"]
          Builtins.foreach(services) do |service|
            NovellUtils.SystemdStartorRestartService(service,"no")
          end
        end


        is_cis_agent_enabled = Service.Enabled("oes-cis-agent")
        Builtins.y2milestone("oes-cis-agent service is enabled : %1",is_cis_agent_enabled)

        if (!is_cis_agent_enabled )
          Service.Enable("oes-cis-agent")
          status = Service.Start("oes-cis-agent")
          Builtins.y2milestone("Nss:Enabled and Started oes-cis-agent service %1,",status)
          NovellUtils.is_cis_agent_enabled = true
       end

        Service.Enable("oes-cis-recall-agent")
        status = Service.Start("oes-cis-recall-agent")
        Builtins.y2milestone("Nss:Enabled and Started oes-cis-recall-agent service : %1",status)

        Service.Enable("oes-cis-scanner")
        status = Service.Start("oes-cis-scanner")
        Builtins.y2milestone("Nss:Enabled and Started oes-cis-scanner service : %1",status)

        # open port 6901 for DFS
        NovellUtils.SetFirewallPort(6901, :tcp, true)

        #in case this is a post install we need to make sure that SMS and the firewall get restarted
        if Mode.normal || Mode.installation && Stage.normal
          Service.Restart("novell-smdrd") if @xadTestMode == false
          Builtins.y2milestone("Nss:restarted novell-smdrd")
          NovellUtils.ReloadFirewall
        end
      end

      # Setting nit ranges
      NitRangeSettings()

      if @xadTestMode == false
        Service.Enable("novell-nit")
        Service.Restart("novell-nit")
        Builtins.y2milestone("novell-nit service restarted")
      end

      Builtins.y2milestone("Nss:writing and flushing SERVICE_CONFIGURED: yes")
      SCR.Write(path(".sysconfig.nss.SERVICE_CONFIGURED"), "yes")
      SCR.Write(path(".sysconfig.nss.SERVICE_PREVIOUSLY_CONFIGURED"), "yes")
      SCR.Write(path(".sysconfig.nss"), nil) # flush the file
      Builtins.sleep(3000) #sleep 3 seconds

      Progress.Finish
      return false if Abort()
      true
    end

    # Get all nss settings from the first parameter
    # (For use by autoinstallation.)
    # @param [Hash] settings The YCP structure to be imported.
    # @return [Boolean] True on success
    def Import(settings)
      settings = deep_copy(settings)
      if settings == {}
        @globals_configured = false
        Ops.set(NovellAddonInstall.products_to_configure, "nss", false)
        return true
      end

      @doRecalculate = false # don't recalculate these values in Write() since they are set by the autoYaST control file.
      Ops.set(
        @Settings,
        "CONFIG_NSS_LDAP_SERVER",
        Ops.get_string(settings, "ldap_server", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_NSS_NSSADMIN_DN",
        Ops.get_string(settings, "nssadmin_dn", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_NSS_EDIR_SERVER_CONTEXT",
        Ops.get_string(settings, "nss_edir_context", "")
      )
      Ops.set(
        @Settings_nit,
        "CONFIG_NIT_START_RANGE",
        Ops.get_integer(settings, "nit_start_range", 100000)
      )
      Ops.set(
        @Settings_nit,
        "CONFIG_NIT_END_RANGE",
        Ops.get_integer(settings, "nit_end_range", 200000)
      )

      @globals_configured = true
      true
    end

    # Dump the nss settings to a single map
    # (For use by autoinstallation.)
    # @return [Hash] Dumped settings (later acceptable by Import ())
    def Export
      settings = {}

      return settings if NovellUtils.CloneSettingstoAutoinstXml("novell-nss") == false

      if @auto_clone
        Ops.set(NovellAddonInstall.products_to_configure, "nss", true)
      end
      if Ops.get(NovellAddonInstall.products_to_configure, "nss", true)
        Ops.set(
          settings,
          "ldap_server",
          Ops.get_string(@Settings, "CONFIG_NSS_LDAP_SERVER", "")
        )
        Ops.set(
          settings,
          "nssadmin_dn",
          Ops.get_string(@Settings, "CONFIG_NSS_NSSADMIN_DN", "")
        )
        Ops.set(
          settings,
          "nss_edir_context",
          Ops.get_string(@Settings, "CONFIG_NSS_EDIR_SERVER_CONTEXT", "")
        )
        Ops.set(
          settings,
          "nit_start_range",
          Ops.get_integer(@Settings_nit, "CONFIG_NIT_START_RANGE", 100000)
        )
        Ops.set(
          settings,
          "nit_end_range",
          Ops.get_integer(@Settings_nit, "CONFIG_NIT_END_RANGE", 200000)
        )
      end
      deep_copy(settings)
    end

    # Create a textual summary and a list of unconfigured cards
    # @return summary of the current configuration
    def Summary
      summary = ""

      if !@globals_configured
        nc = Summary.NotConfigured
        summary = Summary.AddLine(summary, nc)
      else
        summary = Summary.OpenList(summary)

        if Mode.update || Mode.autoupgrade
          summary = Summary.AddListItem(
            summary,
            _("Upgrading from earlier version of OES: yes")
          )
        end
        summary = Summary.AddListItem(
          summary,
          Builtins.sformat(
            _("LDAP Server: %1"),
            Ops.get_string(@Settings, "CONFIG_NSS_LDAP_SERVER", "")
          )
        )
        summary = Summary.AddListItem(
          summary,
          Builtins.sformat(
            _("NSS Admin name: %1"),
            Ops.get_string(@Settings, "CONFIG_NSS_NSSADMIN_DN", "")
          )
        )
        summary = Summary.AddListItem(
          summary,
          Builtins.sformat(
            _("UID start range: %1"),
            Ops.get_integer(@Settings_nit, "CONFIG_NIT_START_RANGE", 100000)
          )
        )

        summary = Summary.AddListItem(
          summary,
          Builtins.sformat(
            _("UID end range: %1"),
            Ops.get_integer(@Settings_nit, "CONFIG_NIT_END_RANGE", 200000)
          )
        )

        summary = Summary.CloseList(summary)
      end
      [summary, []]
    end

    # Create an overview table with all configured cards
    # @return table items
    def Overview
      # TODO FIXME: your code here...
      []
    end


    # Return packages needed to be installed and removed during
    # Autoinstallation to insure module has all needed software
    # installed.
    # @return [Hash] with 2 lists.
    def AutoPackages
      # TODO FIXME: your code here...
      { "install" => [], "remove" => [] }
    end

    publish :variable => :auto_clone, :type => "boolean"
    publish :variable => :Settings, :type => "map"
    publish :variable => :Settings_nit, :type => "map"
    publish :variable => :oldSettings, :type => "map"
    publish :variable => :edir_service_configured, :type => "boolean"
    publish :variable => :schemaLogPath, :type => "string"
    publish :variable => :oesContext, :type => "string"
    publish :variable => :userChangedData, :type => "boolean"
    publish :variable => :doRecalculate, :type => "boolean"
    publish :variable => :xad_packages_installed, :type => "boolean"
    publish :variable => :xadMode, :type => "boolean"
    publish :variable => :xadTestMode, :type => "boolean"
    publish :function => :Modified, :type => "boolean ()"
    publish :variable => :modified, :type => "boolean"
    publish :variable => :proposal_valid, :type => "boolean"
    publish :variable => :reinstallMode, :type => "boolean"
    publish :variable => :write_only, :type => "boolean"
    publish :variable => :globals_configured, :type => "boolean"
    publish :variable => :AbortFunction, :type => "boolean ()"
    publish :function => :Abort, :type => "boolean ()"
    publish :function => :GetOrCreateGID, :type => "string (string)"
    publish :function => :GetOrCreateUID, :type => "string (string, string)"
    publish :function => :ShutDownNamcd, :type => "boolean ()"
    publish :function => :DeleteXtierLocals, :type => "boolean ()"
    publish :function => :DeleteLocals, :type => "boolean ()"
    publish :function => :GetGIDfromLUM, :type => "string (string)"
    publish :function => :GetUIDfromLUM, :type => "string (string)"
    publish :function => :GenerateScriptForFindFilesAndChangeGIDandUID, :type => "string (string, string, string, string)"
    publish :function => :WriteScriptFileForChown, :type => "boolean (string, string, string, string)"
    publish :function => :AppendScriptFileForXtierChown, :type => "boolean (string, string, string, string, string, string)"
    publish :function => :NitRangeDialog, :type => "any ()"
    publish :function => :NitRangeNextButton, :type => "any ()"
    publish :function => :NitReadSettings, :type => "void ()"
    publish :function => :NitRangeSettings, :type => "void ()"
    publish :function => :Read, :type => "boolean ()"
    publish :function => :ChangeModForNssidFile, :type => "boolean ()"
    publish :function => :EdirRetryDone, :type => "boolean ()"
    publish :function => :Write, :type => "boolean ()"
    publish :function => :Import, :type => "boolean (map)"
    publish :function => :Export, :type => "map ()"
    publish :function => :Summary, :type => "list ()"
    publish :function => :Overview, :type => "list ()"
    publish :function => :AutoPackages, :type => "map ()"
  end

  Nss = NssClass.new
  Nss.main
end
